/************************************************************
  FontTaffer - Thief 3 font creation program
  by Crispy (updated by snobel)

"Taffing around with Thief 3 fonts so you don't have to..."

A simple command-line program that takes a TTF or FON
file (TrueType and raster fonts, respectively) and
creates a CEL text file and BMP image file. Once the
BMP file has been converted into a DDS texture file,
Thief 3 will be able to use the font.

Written in the Digital Mars D programming language.
See http://digitalmars.com/d/

************************************************************/

string progVersion = "0.7.0";

private {
  import derelict.sdl2.sdl;
  import derelict.sdl2.ttf;
  import derelict.util.exception;
  import std.getopt;
  import std.file;
  import std.path;
  import std.stdio;
  import std.string;
  import std.conv;
  import std.typecons;
}
pragma(lib, "DerelictUtil.lib");
pragma(lib, "DerelictSDL2.lib");

version (Windows) {
  import std.c.windows.windows;
  pragma(lib, "Kernel32.lib");
}

int abs(int i) {
  if (i<0) return -i;
  else return i;
}

int max(int a, int b) {
  return a > b ? a : b;
}

char safe_char(ushort c) {
  return to!char(c >= 32 && c < 127 || c >= 160 && c < 256 ? c : ' ');
}

// The arg must be left trimmed of whitespace
ushort getCharCode(const string s) {
  if (s.length == 3 && s[0] == '\'' && s[2] == '\'') {
      // Char in single quotes
      return to!ushort(s[1]);
  }
  if (s.length >= 3 && s[0] == '0' && (s[1] == 'x' || s[1] == 'X')) {
    // Hex value
    return toImpl!ushort(s[2..s.length], 16);
  }
  // Decimal value
  return to!ushort(s);
}

int main(string[] args) {

  //// Parse arguments
  
  string inputfilename = "";
  bool help = false;
  string outputfilename = "";
  uint fontsize = 36;
  uint texturesize = 512;
  float linespacing = 1.0;
  long fontindex = 0;
  bool fontbold = false, fontboldish = false, fontitalic = false;
  int offset_left = 0, offset_top = 0;
  int add_width = 0; int add_height = 0;
  bool solid = false, shaded = false;
  ushort padding = 1;
  string adj_filename = "";
  string oldcel_filename = "";
  bool outline_glyphs = false;
  bool render_negx = false;
  bool keep_high = false, keep_low = false;
  ushort charlow = 0, charhigh = 255;
  
  getopt(args, std.getopt.config.passThrough,
    "help|h",           &help,
    "output|o",         &outputfilename,
    "size|s",           &fontsize,
    "texture-size|t",   &texturesize,
    "index|x",          &fontindex,
    "bold|b",           &fontbold,
    "boldish|d",        &fontboldish,
    "italic|i",         &fontitalic,
    "offset-left|u",    &offset_left,
    "offset-top|v",     &offset_top,
    "add-width|w",      &add_width,
    "add-height|e",     &add_height,
    "line-spacing|l",   &linespacing,
    "padding|p",        &padding,
    "render-negx|r",    &render_negx,
    "outline-glyphs|g", &outline_glyphs,
    "adj-file|a",       &adj_filename,
    "oldcel-file|c",    &oldcel_filename,
    "solid",            &solid,
    "shaded",           &shaded,
    "keep-low",         &keep_low,
    "keep-high",        &keep_high,
    "first",            &charlow,
    "last",             &charhigh
  );
  
  if (help) {
    showhelp();
    return 0;
  }          

  // Remaining args should be the program name and the font file
  if (args.length < 2) {
    writefln("Error: Please specify a font file to read.");
    return 1;
  } 
  else {
    foreach (string arg; args[1..args.length]) {
      // Assume the first non-option arg is the font file
      if (arg[0] != '-' && !inputfilename.length) {
        inputfilename = arg;
      }
      else {
        writefln("Error: Unrecognised option '%s'", arg);
        return 9;
      }
    }
  }

  // Input filename is required...
  if (!inputfilename.length) {
    writefln("Error: Please specify a font file to read.");
    return 1;
  }
  // ...and defaults to a TTF file
  inputfilename = defaultExtension(inputfilename, "ttf");
  // If the file is not found locally, try the system fonts dir
  version (Windows) {
    if (! exists(inputfilename)) {
      wchar[MAX_PATH] win_dir;
      uint len = GetEnvironmentVariableW("WINDIR", &win_dir[0], MAX_PATH);
      if (! len) {
        writefln("Error: The system fonts dir could not be read.");
        return 12;
      }
      inputfilename = to!string(win_dir[0..len]) ~ "\\Fonts\\" ~ baseName(inputfilename);
    }
  }

  // Output filename defaults to input filename (but with different extensions, and no path)
  if (!outputfilename.length) {
    outputfilename = baseName(inputfilename);
  }
  string outputcelfilename = setExtension(outputfilename, "cel");
  outputfilename = setExtension(outputfilename, "bmp");
  
  // Structure for individual glyph adjustments. Initializes to all zeroes, i.e. no change
  struct Glyph_adj {
    int offs_l, offs_t;
    int add_w;
  }
  Glyph_adj[256] glyph_tuning;  // (What is the D equivalent of map<ushort, Glyph_adj>?)

  // If there is a file for individual glyph adjustments, parse it now to catch any errors
  if (adj_filename.length) {
    adj_filename = defaultExtension(adj_filename, "adj");  
    alias Tuple!(string, int, int, int, string) Glyph_tuple;
    try {
      Glyph_tuple[] tuples = slurp!Glyph_tuple(adj_filename, " %s %s %s %s %s");
      foreach (Glyph_tuple adj; tuples) {
        ushort code = getCharCode(adj[0]);
        glyph_tuning[code].offs_l = adj[1];
        glyph_tuning[code].offs_t = adj[2];
        glyph_tuning[code].add_w = adj[3];
        debug writefln("Character %s (%d) has extra tuning: offset_left %d, offset_top %d, add_width %d", 
                       safe_char(code), code, adj[1], adj[2], adj[3]);
      }
    }
    // Error handling could be more detailed but the message provided by the exception is 
    // a bit too scary to throw at the user.
    catch (Exception e) {
      writeln("Error: Glyph adjustment file trouble. (Missing file or formatting errors?)");
      return 10;
    }
  }
  
  // Old or new CEL data will be stored here
  int[] celdata;
  int   lineheight = 0;
  
  // If there is an old CEL file, parse it now to catch any errors
  if (oldcel_filename.length) {
    oldcel_filename = defaultExtension(oldcel_filename, "cel");
    try {
      // Read and remove the commas and control chars
      string cel_ln = readText(oldcel_filename);
      cel_ln = tr(cel_ln, ",\t\r\n", " ", "");
      debug writefln("Old CEL file contains %s", cel_ln);
      string[] entries = split(cel_ln);
      // Convert to an int array
      foreach (string s; entries) {
        munch(s, " ");
        // Stop if a SU scaling factor entry is found
        if (s[0] == '*') break;
        celdata ~= to!int(s);
      }
      debug writefln("Old CEL file has %d entries", celdata.length);
    }
    // Error handling could be more detailed etc.
    catch (Exception e) {
      writeln("Error: Old CEL file trouble. (Missing file or formatting errors?)");
      return 10;
    }
    if (celdata.length < 2) {
      writeln("Error: Old CEL file contains too few entries");
      return 11;
    }
    padding = to!ushort(celdata[0]);
    lineheight = to!ushort(celdata[1]);
  }
  
  // For glyph outlines there must be some padding
  if (outline_glyphs) {
    padding = to!ushort(max(1, padding));
  }
  
  //// Set up the font
  
  DerelictSDL2.load();
  DerelictSDL2ttf.load();
  if(TTF_Init()<0) {
      writefln("Failed to initialise font library: "~to!string(TTF_GetError()));
      return 2;
  }
  
  TTF_Font* font = TTF_OpenFontIndex(.toStringz(inputfilename), fontsize, to!int(fontindex));
  
  if (!font) {
    writefln("Failed to open %s (size %s): %s", inputfilename, fontsize, to!string(TTF_GetError()));
    if (fontindex>0) {
      writefln("Perhaps the file does not contain this many fonts; try a lower-numbered font\nindex.");
    }
    TTF_Quit();
    return 3;
  }
  
  debug {
    writefln("Input file: %s, CEL file: %s, BMP file: %s, font size: %d",
      inputfilename, outputcelfilename, outputfilename, fontsize);
  }
  
  // Set font style
  int style=TTF_STYLE_NORMAL;
  if (fontbold) style |= TTF_STYLE_BOLD;
  if (fontitalic) style |= TTF_STYLE_ITALIC;
  TTF_SetFontStyle(font, style);
  
  //// Write CEL file
   
  if (! celdata.length) {
  
    // First number is the size of the padding above each line and to the left of each character
    celdata ~= padding;
    
    // Second number is height of a line (without padding, but with the global adjustment)
    lineheight = max(1, to!int(TTF_FontHeight(font) * linespacing) + add_height);
    if (fontboldish) {
      // Needs 2 added pixels even if the glyph is only rendered twice vertically?!
      //++lineheight;
      lineheight += 2;
    }
    celdata ~= lineheight;
    
    // Now we calculate character data for the CEL file.
    // The CEL file lists characters encoded in Latin-1 format (0 to 0xFF).
    // Thankfully, Latin-1 is exactly what SDL_ttf accepts! So we just
    // loop from 0 to 0xFF (255) and ask SDL_ttf how wide the character is.
    int charwidth, xmin, xmax;
    int linewidth = 0;
    
    for (ushort c = 0; c <= 255; ++c) {
      if (c < charlow || c > charhigh || (!keep_low && c < 32) || (!keep_high && c >= 127 && c < 160) || 
          TTF_GlyphMetrics(font, c, &xmin, &xmax, null, null, null) < 0) {
        // Failed; couldn't render that glyph, so put in a 0 entry (meaning skip it).
        // Note that values below 32 (space character) are automatically skipped;
        // even if they're in the font, we don't want them. By default the same goes 
        // for the range 127 to 159, which is unused in the original fonts
        celdata ~= 0;
        debug writefln("Character %s (%d) was skipped", safe_char(c), c);
      } else {
        
        // Drop the part of the glyph that 'goes backwards' on the line (negative x)
        // unless overruled by the user
        if (xmin < 0 && ! render_negx) {
          charwidth = xmax;
        }
        else {
          charwidth = xmax - xmin;
        }
        
        // Apply the global and individual width adjustment but make sure the character 
        // has a positive width
        charwidth = max(1, charwidth + add_width + glyph_tuning[c].add_w);
        if (fontboldish) {
          charwidth += 2;
        }
        
        debug writefln("Character %s (%d) has xmin %d, xmax %d - width %d", 
                       safe_char(c), c, xmin, xmax, charwidth);
        
        // The width of this character is in charwidth
        if (linewidth + 2 * padding + charwidth > texturesize) {
          // Too long! Wrap to next line.
          celdata ~= -1; // Add a line break
          linewidth = 0; // The next line has no width yet
          debug writefln("Wrapping before character %d", c);
        }
                            
        // Add character to line
        linewidth += padding + charwidth;
        celdata ~= charwidth;
      }
    }

  }
  
  // Write to file
  string celtext;
  for (int i=0; i<celdata.length; i++) {
    if (i!=0) celtext ~= ", ";
    celtext ~= to!string(celdata[i]);
  }
  std.file.write(outputcelfilename, celtext);

  //// Write BMP file
  
  // Initialise SDL because we're going to be using some of its functions
  if (SDL_Init(SDL_INIT_VIDEO) < 0) {
    writefln("Failed to initialise SDL: ", to!string(SDL_GetError()));
    return 4;
  }
  
  // Create a surface to write to
  SDL_Surface* bitmap = SDL_CreateRGBSurface(SDL_SWSURFACE, texturesize, texturesize, 32, 0, 0, 0, 0);
  if (!bitmap) {
    writefln("CreateRGBSurface failed: ", to!string(SDL_GetError()));
    return 6;
  }

  // Optionally fill the surface with red so that the padding and glyph outlines will be visible.
  if (outline_glyphs && SDL_FillRect(bitmap, null, SDL_MapRGB(bitmap.format, 255, 0, 0))) {
    writefln("FillRect for glyph outlines failed: ", to!string(SDL_GetError()));
    return 5;
  }
  
  // Write glyphs to the surface
  int destx = padding, desty = padding;
  
  SDL_Color color, bg;
  color.r = color.g = color.b = 255;
  bg.r = bg.g = bg.b = 0;
  
  int i = -1;
  
  foreach (int cel; celdata[2..celdata.length]) {
    
    // Increment glyph number
    i++;
    
    if (cel == -1) {
    
      // Skip line, don't increment glyph number
      destx = padding;
      desty += lineheight + padding;
      
      // Undo above increment
      i--;
      
    } else if (cel > 0) {

      // Get position to render glyph at
      int minx = 0;
      if (TTF_GlyphMetrics(font, to!ushort(i), &minx, null, null, null, null) <0 ) {
        writefln("Warning: Failed to get metrics of character %d", i);
      }
      
      // Render this glyph
      SDL_Surface* glyph = null;
      if (TTF_GlyphIsProvided(font, to!ushort(i))) {
        if (solid) {
          glyph = TTF_RenderGlyph_Solid(font, to!ushort(i), color);
        }
        else if (shaded) {
          glyph = TTF_RenderGlyph_Shaded(font, to!ushort(i), color, bg);
        }
        else {
          glyph = TTF_RenderGlyph_Blended(font, to!ushort(i), color);
        }
      }
      
      SDL_Rect srcrect, destrect;
      
      // Drop the part of the glyph that 'goes backwards' on the line (negative x)
      // unless overruled by the user
      if (render_negx) {
        srcrect.x = max(0, minx);
      }
      else {
        srcrect.x = abs(minx);
      }
      // Apply the offsets to the source rect to avoid drawing into other glyphs
      srcrect.x -= offset_left + glyph_tuning[i].offs_l;
      srcrect.y = - (offset_top + glyph_tuning[i].offs_t);
      srcrect.w = cel;
      srcrect.h = lineheight;
      destrect.x = to!short(destx);
      destrect.y = to!short(desty);

      // For the outline to show in non-shading modes the background must be filled in
      if (outline_glyphs) {
        destrect.w = srcrect.w;
        destrect.h = srcrect.h;
        if (SDL_FillRect(bitmap, &destrect, SDL_MapRGB(bitmap.format, 0, 0, 0))) {
          writefln("FillRect failed: ", to!string(SDL_GetError()));
          return 8;
        }
      }

      if (glyph) {
        // Ugly fake bold mode because SDL's bold does not work well for T3
        int x_pix = 1, y_pix = 1;
        if (fontboldish) {
          x_pix = 3; y_pix = 2;
        }
        for (int x = 0; x < x_pix; ++x) {
          SDL_Rect drect = destrect;
          drect.x += x;
          SDL_Rect srect = srcrect;
          srect.w -= x;
          if (SDL_BlitSurface(glyph, &srect, bitmap, &drect)<0) {
            writefln("Warning: Blit of character %d failed", i);
          }
        }
        for (int y = 0; y < y_pix; ++y) {
          SDL_Rect drect = destrect;
          drect.x += x_pix / 2; drect.y += y;
          SDL_Rect srect = srcrect;
          srect.h -= y;
          if (SDL_BlitSurface(glyph, &srect, bitmap, &drect)<0) {
            writefln("Warning: Blit of character %d failed", i);
          }
        }
      }
      // Move across for next glyph
      destx += cel + padding;
    }
    
  }
  
  // Save surface
  if (SDL_SaveBMP(bitmap, toStringz(outputfilename))) {
    writefln("SaveBMP failed: %s", to!string(SDL_GetError()));
    return 7;
  }

  //// Clean up

  TTF_Quit();
  SDL_Quit();
  
  return 0;
}

void showhelp() {
  writeln("");
  writefln("FontTaffer v%s by Crispy (updates by snobel)", progVersion);
  writeln("");
  writefln("Usage: fonttaff [options] truetype.ttf");
  writefln("       fonttaff [options] raster.fon");
  writeln("");
  writefln("   -h, --help                 Displays this message");
  writefln("   -o, --output=FILENAME      Output filename (optional)");
  writefln("   -s, --size=FONTSIZE        Font size to use (default 36)");
  writefln("   -t, --texture-size=PIXELS  Size of output texture (default 512)");
  writefln("   -l, --line-spacing=FACTOR  Line spacing (default 1.0)");
  writefln("   -d, --boldish              Render the font in pseudo-bold. Experimental, may");
  writefln("                              work better than bold for thin horizontal lines");
  writefln("   -b, --bold                 Render the font in bold type");
  writefln("   -i, --italic               Render the font in italics");
  writefln("   -x, --index=FONTINDEX      Font index (default 0). You will");
  writefln("                              usually not have to change this");
  writefln("   -r, --render-negx          Render the full glyphs, including parts that");
  writefln("                              would normally 'go backwards' on the line");
  writefln("   -u, --offset-left=PIXELS   Shift all glyphs PIXELS pixels to the right");
  writefln("                              (default 0). See note below");
  writefln("   -v, --offset-top=PIXELS    Shift all glyphs PIXELS pixels downwards");
  writefln("                              (default 0). See note below");
  writefln("   -w, --add-width=PIXELS     Add PIXELS pixels on the right side of each glyph");
  writefln("                              (default 0). See note below");
  writefln("   -e, --add-height=PIXELS    Add PIXELS pixels on the bottom of each glyph");
  writefln("                              (default 0). See note below");
  writefln("   -a, --adj-file=FILENAME    File containing individual glyph adjustments.");
  writefln("                              See note below");
  writefln("   -c, --oldcel-file=FILENAME Use padding, line height and character widths");
  writefln("                              from an existing .cel file instead of calculating");
  writefln("   -p, --padding=PIXELS       Number of blank pixels above each row and");
  writefln("                              before each glyph (default 1)");
  writefln("   -g, --outline-glyphs       Render the glyphs on a red background, to better");
  writefln("                              show the size of each. Implies --padding");
  writefln("       --solid                Render using the 'solid' mode (no anti-aliasing)");
  writefln("       --shaded               Render using the 'shaded' mode (not recommended)");
  writefln("       --keep-low             Include characters 0-31");
  writefln("       --keep-high            Include characters 127-159");
  writefln("       --first=NUM            Start at this character (default 0)");
  writefln("       --last=NUM             Stop at this character (default 255)");
  writeln("");
  writefln("Note:  No wildcards are allowed in filenames.");
  writeln("");
  writefln("Note:  Use the --offset-left and --offset-right options to adjust the placement");
  writefln("       of the glyphs. Use the --add-width and --add-height options to adjust ");
  writefln("       the size of the glyphs. This will change the spacing, not the size of");
  writefln("       the rendered characters. Negative values are allowed for all 4 options.");
  writefln("       In addition to these adjustments which affect all characters, some may");
  writefln("       need individual tuning. Use the --adj-file option to specify a file");
  writefln("       containing lines of the following format: ");
  writefln("               <code> <offset-left> <offset-top> <add-width> [<comment>]");
  writefln("       The <code> field can be decimal, hex (e.g. 0xA0) or a character (except");
  writefln("       space) in single quotes. The values will be applied on top of the global");
  writefln("       values for the specified character codes. Example: \"32 0 0 6 ; SPC\" will");
  writefln("       widen SPACE by 6 pixels.");
  writefln("       In all cases only small values make sense but adjustments for one glyph");
  writefln("       will not affect the rendering of neighbouring glyphs. To get an idea of");
  writefln("       how the characters will look, use the --outline-glyphs option.");
  writeln("");
  writefln("Thanks to New Horizon and everyone else at TTLG!");
  writefln("This program is based in part on the work of the FreeType team");
  writefln("(http://www.freetype.org)");
}
